@@ -12,6 +12,7 @@ from click.testing import CliRunner
 
 import libtmux
 from libtmux.common import has_lt_version
+from libtmux.exc import LibTmuxException
 from tmuxp import cli, config
 from tmuxp.cli import (
     command_ls,
@@ -406,6 +407,262 @@ def test_load_zsh_autotitle_warning(cli_args, tmpdir, monkeypatch):
         assert 'Please set' not in result.output
 
 
+@pytest.mark.parametrize("cli_cmd", ['shell', 'shell_plus'])
+@pytest.mark.parametrize(
+    "cli_args,inputs,env,expected_output",
+    [
+        (
+            ['-L{SOCKET_NAME}', '-c', 'print(str(server.socket_name))'],
+            [],
+            {},
+            '{SERVER_SOCKET_NAME}',
+        ),
+        (
+            [
+                '-L{SOCKET_NAME}',
+                '{SESSION_NAME}',
+                '-c',
+                'print(session.name)',
+            ],
+            [],
+            {},
+            '{SESSION_NAME}',
+        ),
+        (
+            [
+                '-L{SOCKET_NAME}',
+                '{SESSION_NAME}',
+                '{WINDOW_NAME}',
+                '-c',
+                'print(server.has_session(session.name))',
+            ],
+            [],
+            {},
+            'True',
+        ),
+        (
+            [
+                '-L{SOCKET_NAME}',
+                '{SESSION_NAME}',
+                '{WINDOW_NAME}',
+                '-c',
+                'print(window.name)',
+            ],
+            [],
+            {},
+            '{WINDOW_NAME}',
+        ),
+        (
+            [
+                '-L{SOCKET_NAME}',
+                '{SESSION_NAME}',
+                '{WINDOW_NAME}',
+                '-c',
+                'print(pane.id)',
+            ],
+            [],
+            {},
+            '{PANE_ID}',
+        ),
+        (
+            [
+                '-L{SOCKET_NAME}',
+                '-c',
+                'print(pane.id)',
+            ],
+            [],
+            {'TMUX_PANE': '{PANE_ID}'},
+            '{PANE_ID}',
+        ),
+    ],
+)
+def test_shell(
+    cli_cmd,
+    cli_args,
+    inputs,
+    expected_output,
+    env,
+    tmpdir,
+    monkeypatch,
+    server,
+    session,
+):
+    monkeypatch.setenv('HOME', str(tmpdir))
+    window_name = 'my_window'
+    window = session.new_window(window_name=window_name)
+    window.split_window()
+
+    template_ctx = dict(
+        SOCKET_NAME=server.socket_name,
+        SOCKET_PATH=server.socket_path,
+        SESSION_NAME=session.name,
+        WINDOW_NAME=window_name,
+        PANE_ID=window.attached_pane.id,
+        SERVER_SOCKET_NAME=server.socket_name,
+    )
+
+    cli_args = [cli_cmd] + [cli_arg.format(**template_ctx) for cli_arg in cli_args]
+
+    for k, v in env.items():
+        monkeypatch.setenv(k, v.format(**template_ctx))
+
+    with tmpdir.as_cwd():
+        runner = CliRunner()
+
+        result = runner.invoke(
+            cli.cli, cli_args, input=''.join(inputs), catch_exceptions=False
+        )
+        assert expected_output.format(**template_ctx) in result.output
+
+
+@pytest.mark.parametrize("cli_cmd", ['shell', 'shell_plus'])
+@pytest.mark.parametrize(
+    "cli_args,inputs,env,template_ctx,exception,message",
+    [
+        (
+            ['-LDoesNotExist', '-c', 'print(str(server.socket_name))'],
+            [],
+            {},
+            {},
+            LibTmuxException,
+            r'.*DoesNotExist\s\(No such file or directory\).*',
+        ),
+        (
+            [
+                '-L{SOCKET_NAME}',
+                'nonexistant_session',
+                '-c',
+                'print(str(server.socket_name))',
+            ],
+            [],
+            {},
+            {'session_name': 'nonexistant_session'},
+            None,
+            'Session not found: nonexistant_session',
+        ),
+        (
+            [
+                '-L{SOCKET_NAME}',
+                '{SESSION_NAME}',
+                'nonexistant_window',
+                '-c',
+                'print(str(server.socket_name))',
+            ],
+            [],
+            {},
+            {'window_name': 'nonexistant_window'},
+            None,
+            'Window not found: {WINDOW_NAME}',
+        ),
+    ],
+)
+def test_shell_target_missing(
+    cli_cmd,
+    cli_args,
+    inputs,
+    env,
+    template_ctx,
+    exception,
+    message,
+    tmpdir,
+    monkeypatch,
+    socket_name,
+    server,
+    session,
+):
+    monkeypatch.setenv('HOME', str(tmpdir))
+    window_name = 'my_window'
+    window = session.new_window(window_name=window_name)
+    window.split_window()
+
+    template_ctx = dict(
+        SOCKET_NAME=server.socket_name,
+        SOCKET_PATH=server.socket_path,
+        SESSION_NAME=session.name,
+        WINDOW_NAME=template_ctx.get('window_name', window_name),
+        PANE_ID=template_ctx.get('pane_id'),
+        SERVER_SOCKET_NAME=server.socket_name,
+    )
+    cli_args = [cli_cmd] + [cli_arg.format(**template_ctx) for cli_arg in cli_args]
+
+    for k, v in env.items():
+        monkeypatch.setenv(k, v.format(**template_ctx))
+
+    with tmpdir.as_cwd():
+        runner = CliRunner()
+
+        if exception is not None:
+            with pytest.raises(exception, match=message.format(**template_ctx)):
+                result = runner.invoke(
+                    cli.cli, cli_args, input=''.join(inputs), catch_exceptions=False
+                )
+        else:
+            result = runner.invoke(
+                cli.cli, cli_args, input=''.join(inputs), catch_exceptions=False
+            )
+            assert message.format(**template_ctx) in result.output
+
+
+@pytest.mark.parametrize(
+    "cli_args,inputs,env,message",
+    [
+        (
+            [
+                'shell_plus',
+                '-L{SOCKET_NAME}',
+            ],
+            [],
+            {},
+            '(InteractiveConsole)',
+        ),
+        (
+            [
+                'shell_plus',
+                '-L{SOCKET_NAME}',
+            ],
+            [],
+            {'PANE_ID': '{PANE_ID}'},
+            '(InteractiveConsole)',
+        ),
+    ],
+)
+def test_shell_plus(
+    cli_args,
+    inputs,
+    env,
+    message,
+    tmpdir,
+    monkeypatch,
+    server,
+    session,
+):
+    monkeypatch.setenv('HOME', str(tmpdir))
+    window_name = 'my_window'
+    window = session.new_window(window_name=window_name)
+    window.split_window()
+
+    template_ctx = dict(
+        SOCKET_NAME=server.socket_name,
+        SOCKET_PATH=server.socket_path,
+        SESSION_NAME=session.name,
+        WINDOW_NAME=window_name,
+        PANE_ID=window.attached_pane.id,
+        SERVER_SOCKET_NAME=server.socket_name,
+    )
+
+    cli_args[:] = [cli_arg.format(**template_ctx) for cli_arg in cli_args]
+    for k, v in env.items():
+        monkeypatch.setenv(k, v.format(**template_ctx))
+
+    with tmpdir.as_cwd():
+        runner = CliRunner()
+
+        result = runner.invoke(
+            cli.cli, cli_args, input=''.join(inputs), catch_exceptions=True
+        )
+        assert message.format(**template_ctx) in result.output
+
+
 @pytest.mark.parametrize(
     "cli_args",
     [
